function [values] = basis_eval(basis_types, basis_indices, nodes)
%BASIS_EVAL evaluates basis functions for BOPB building full matrix.
%  Supports Fourier, (preconditioned) Legendre, and Chebyshev basis
%    (also mixed bases).
%
%  Input:
%    basis_types - cell array of length d,
%                  contains either 'fourier' or 'chebyshev' or 'legendre'
%    basis_indices - double matrix: num_indices \times d
%    nodes - spatial evaluation nodes, double matrix: num_nodes \times d
%
%  Output:
%    values - double matrix: num_nodes \times num_indices
%             containing BOPB with indices evaluated at nodes
%
% version with speed improvements and generalizations 2018-2019 by Toni Volkmer

d = size(nodes, 2);
if d ~= size(basis_indices, 2)
  error('dimension mismatch basis_indices and nodes');
end
if length(basis_types) ~= d
  error('dimension mismatch basis_indices and basis_types');
end
num_nodes = size(nodes, 1);
num_indices = size(basis_indices, 1);

values = ones(num_nodes, num_indices);

for j = 1:d
  basis_indices_j = basis_indices(:,j);
  nodes_j = nodes(:,j);
  basis_type_j = basis_types{j};
  switch basis_type_j
    case 'fourier'
      basis_indices_j_transformed = (-1).^basis_indices_j.*ceil(basis_indices_j/2);
      values = values .* exp(2*pi*1i * nodes_j * basis_indices_j_transformed.');
    case 'chebyshev'
      values = values .* repmat(2.^((basis_indices_j.'~=0)/2),num_nodes,1) .* cos(acos(nodes_j) * basis_indices_j.');
    case {'legendre', 'legendre_preconditioned'}
      max_basis_ind = max(basis_indices_j);
      eval_table = zeros(length(nodes_j), max_basis_ind + 1);
      eval_table(:,1) = 1;
      if max_basis_ind > 0
        eval_table(:,2) = nodes_j;
      end
      for cur_n = 2:max_basis_ind
        eval_table(:,cur_n+1) = (2*cur_n-1)/cur_n * nodes_j .* eval_table(:,cur_n) - (cur_n-1)/cur_n * eval_table(:,cur_n-1);
      end
      for cur_n = 0:max_basis_ind
        eval_table(:,cur_n+1) = sqrt((2*cur_n+1)) * eval_table(:,cur_n+1);
      end
      if strcmp(basis_type_j, 'legendre_preconditioned')
        for cur_n = 0:max_basis_ind
          eval_table(:,cur_n+1) = (pi/2)^(1/2) * (1-nodes_j.^2).^(1/4) .* eval_table(:,cur_n+1);
        end
      end
      values = values .* eval_table(:,basis_indices_j+1);
%       for ii=1:size(values,2)
%         values(:,ii) = values(:,ii) .* legendreP(basis_indices_j(ii),nodes_j) * sqrt((2*basis_indices_j(ii)+1)/2);
%       end
    otherwise
      error('unknown basis type: %s', basis_type_j);
  end
end

end

