function [num_iter,error_n,error_c,aa,cc,tElapsed_sam,tElapsed_tot,SF,sam_num]=sublin_cosamp_v2_func_fct(test_fct,N,d,basis_types,s,m,L,tau,m_bar,m_1,m_2,cosamp_max_iter,pairing_type)
%SUBLIN_COSAMP_V2_FUNC_FCT helper function for general case.
%
%  Input:
%    test_fct - function handle for test function
%    N - scalar, number of possible basis indices per spatial dimension
%        (side length of cube)
%    d - scalar, spatial dimension
%    basis_types - cell array of length d,
%                  contains either 'fourier' or 'chebyshev' or 'legendre'
%                  or 'legendre_preconditioned'
%    s - scalar, sparsity
%    m - parameter for entry identification,
%        empty array for entryId_useall,
%        array of length 2 for entryId_cs,
%        scalar for entryId_old
%    L - parameter for Gaussian weights of entryId_old
%    tau - scalar >= 0, threshold for entry identification
%    m_bar - number of samples for least squares coefficients computation
%    m_1 - number of prefix samples for pairing steo
%    m_2 - number of postfix samples for pairing step
%    cosamp_max_iter - scalar >= 1, termination criterion for algorithm
%    pairing_type - pairing method, either 'pairing' or 'pairing_large'
%
%  Output:
%    num_iter - number of iterations actually performed
%    error_n - error w.r.t. basis indices
%    error_c - error w.r.t. coefficients
%    aa - ALL basis indices (detected + undetected)
%    cc - ALL coefficients (detected + undetected)
%    output - detected basis indices, double matrix: num_indices \times d,
%             P in Algorithm 3 of the paper
%    coeff_est - estimated coefficients, column vector
%    tElapsed_sam - sampling runtime
%    tElapsed_tot - total runtime
%    SF - 1 for success, 0 for failure
%    sam_num - total number of samples
%
% based on SHT implementation 2018 by Bosu Choi
% version with speed improvements and generalizations 2018-2019 by Toni Volkmer

switch pairing_type
  case 'pairing'
    pairing_fct = @pairing;
  case 'pairing_large'
    pairing_fct = @pairing_large;
  otherwise
    error('unsupported pairing type: %s', pairing_type);
end

samples_entryid=cell(d-1,1); % initialization of samples for entryId
samples_pairing=cell(d-1,1); %initialization of samples for pairing
w=cell(d-1,1); %,m_1); %containing 1 by jj vectors / jj=2,...,d % initialization of sample points W_j's in the paper
z=cell(d-2,1); %,m_2); %containing 1 by d-jj vectors / jj=2,...,d % initialization of sample points Z_j's in the paper
wz=cell(d-1,1); %m_1*m_2,d-1); %containing 1 by d vectors % sampling points W_j x Z_j's in the paper

%start counting time for sampling 
tStart_sam= tic;

%random sample points for the least square solving
nodes_ls = basis_gen_rnd_nodes(basis_types, m_bar);

%samples for least square solving
samples_ls = test_fct(nodes_ls);

if isempty(m)
  entryid_fct = @(N, s, d, basis_types, return_type, funcEval, params) entryId_useall(N, d, basis_types, return_type);
elseif length(m) == 2
  entryid_fct = @(N, s, d, basis_types, return_type, funcEval, param_nodes1d) entryId_cs(N, m(1), m(2), s, d, basis_types, return_type, funcEval, tau, param_nodes1d);
elseif L > 0
  entryid_fct = @(N, s, d, basis_types, return_type, funcEval, gauss) entryId_old(N, m, s, L, d, basis_types, return_type, funcEval, tau, gauss);
elseif L == 0
  entryid_fct = @(N, s, d, basis_types, return_type, funcEval, weights) entryId_quad_new(N, m, s, d, basis_types, return_type, funcEval, tau, weights);
else
  error('m > 0 and L < 0 currently not supported');
end

%nodes for entry id
[nodes_entryid, entryid_fct_params] = entryid_fct(N, 2*s, d, basis_types, 'nodes', [], []);

%samples f for entry identification: d by m by N 
for jj=1:1:d
  samples_entryid{jj} = test_fct(nodes_entryid{jj});
end

%random sample points for pairing
for jj=2:1:d
    %W_j's : cell array of size d-1 by m_1 containing 1 by jj vectors
    w{jj-1} = basis_gen_rnd_nodes(basis_types(1:jj), m_1);
    %Z_j's : cell array of size d-2 by m_2 containing 1 by d-jj vectors 
    z{jj-1} = basis_gen_rnd_nodes(basis_types(jj+1:d), m_2);
    %W_j x Z_j 's: cell array of size m_1m_2 by d-1 containing 1 by d vectors
    %ff = f(W_j x Z_j) : multidimensional array of size m_1 by m_2 by d-1 %samples for pairing      
    wz{jj-1} = zeros(m_1*m_2,d);
    for k=1:m_2
      wz{jj-1}(m_1*(k-1)+1 : m_1*k,:) = [w{jj-1} repmat(z{jj-1}(k,:),m_1,1)];
    end
end
%
for jj=2:1:d
  samples_pairing{jj-1} = reshape(test_fct(wz{jj-1}), m_1, m_2);
end
tElapsed_sam = toc(tStart_sam);%stop counting time for sampling

tStart_tot = tic; % Start counting time for main algorithm
%running sublinearized CoSaMP
[a,supp_a,~,num_iter]=sublin_cosamp_v2(samples_ls, nodes_ls, samples_entryid, nodes_entryid, entryid_fct, entryid_fct_params, samples_pairing, w, wz, pairing_fct, N, s, cosamp_max_iter, basis_types);
tElapsed_tot = toc(tStart_tot);% Stop counting time for main algorithm
error_n=nan;
error_c=nan;
aa=supp_a;
cc=a;
SF=nan;
%total sample number
sam_num = numel(samples_ls) + numel(samples_entryid{d});
for jj=1:d-1
  sam_num = sam_num + numel(samples_entryid{jj}) + numel(samples_pairing{jj});
end
