function [num_iter,error_n,error_c,aa,cc,tElapsed_sam,tElapsed_tot,SF,sam_num]=sublin_cosamp_v2_func_noise(N,d,basis_types,s,m,L,tau,m_bar,m_1,m_2,cosamp_max_iter,sr,pairing_type)
%SUBLIN_COSAMP_V2_FUNC_NOISE helper function for exactly sparse case with noise.
%
%  Input:
%    N - scalar, number of possible basis indices per spatial dimension
%        (side length of cube)
%    d - scalar, spatial dimension
%    basis_types - cell array of length d,
%                  contains either 'fourier' or 'chebyshev' or 'legendre'
%                  or 'legendre_preconditioned'
%    s - scalar, sparsity
%    m - parameter for entry identification,
%        empty array for entryId_useall,
%        array of length 2 for entryId_cs,
%        scalar for entryId_old
%    L - parameter for Gaussian weights of entryId_old
%    tau - scalar >= 0, threshold for entry identification
%    m_bar - number of samples for least squares coefficients computation
%    m_1 - number of prefix samples for pairing steo
%    m_2 - number of postfix samples for pairing step
%    cosamp_max_iter - scalar >= 1, termination criterion for algorithm
%    sr - signal-to-noise ratio
%    pairing_type - pairing method, either 'pairing' or 'pairing_large'
%
%  Output:
%    num_iter - number of iterations actually performed
%    error_n - error w.r.t. basis indices
%    error_c - error w.r.t. coefficients
%    aa - ALL basis indices (detected + undetected)
%    cc - ALL coefficients (detected + undetected)
%    output - detected basis indices, double matrix: num_indices \times d,
%             P in Algorithm 3 of the paper
%    coeff_est - estimated coefficients, column vector
%    tElapsed_sam - sampling runtime
%    tElapsed_tot - total runtime
%    SF - 1 for success, 0 for failure
%    sam_num - total number of samples
%
% based on SHT implementation 2018 by Bosu Choi
% version with speed improvements and generalizations 2018-2019 by Toni Volkmer

switch pairing_type
  case 'pairing'
    pairing_fct = @pairing;
  case 'pairing_large'
    pairing_fct = @pairing_large;
  otherwise
    error('unsupported pairing type: %s', pairing_type);
end

samples_entryid=cell(d-1,1); % initialization of samples for entryId
samples_pairing=cell(d-1,1); %initialization of samples for pairing
w=cell(d-1,1); %,m_1); %containing 1 by jj vectors / jj=2,...,d % initialization of sample points W_j's in the paper
z=cell(d-2,1); %,m_2); %containing 1 by d-jj vectors / jj=2,...,d % initialization of sample points Z_j's in the paper
wz=cell(d-1,1); %m_1*m_2,d-1); %containing 1 by d vectors % sampling points W_j x Z_j's in the paper
%randomly chosen coefficients from {-1,1}
c = (-1).^randi(2,s,1);
%randomly chosen multi_index n : 1 by d by s
n = [];
while size(n,1) < s
  n = unique([n; randi(N,[s d])-1], 'rows');
end
n = n(1:s,:);
cur_rng_state = rng;

%start counting time for sampling 
tStart_sam= tic;

%random sample points for the least square solving
nodes_ls = basis_gen_rnd_nodes(basis_types, m_bar);

%samples for least square solving
basisEval = basis_eval(basis_types, n, nodes_ls);
samples_ls = basisEval * c;

%ADDING NOISE
snorm_f_ls=norm(samples_ls)^2;
n_f_ls=randn(m_bar,1);  % random gaussian noise for samples for least square solving 
if ~isreal(samples_ls)
  n_f_ls = n_f_ls + 1i * randn(m_bar,1);
end
snorm_n_f_ls=norm(n_f_ls)^2;   % squared l2 norm of noise
res_const=snorm_f_ls/snorm_n_f_ls*10^(-sr/10);
n_f_ls=sqrt(res_const)*n_f_ls;
samples_ls=samples_ls+n_f_ls;

if isnan(tau)
  tau = max(norm(samples_ls) * sqrt( 10^(-sr/10) / (1 + 10^(-sr/10)) ), 1e-12 * sqrt(numel(samples_ls))); % threshold used in the main loop to terminate the iteration
end

if isempty(m)
  entryid_fct = @(N, s, d, basis_types, return_type, funcEval, params) entryId_useall(N, d, basis_types, return_type);
elseif length(m) == 2
  entryid_fct = @(N, s, d, basis_types, return_type, funcEval, param_nodes1d) entryId_cs(N, m(1), m(2), s, d, basis_types, return_type, funcEval, tau, param_nodes1d);
elseif L > 0
  entryid_fct = @(N, s, d, basis_types, return_type, funcEval, gauss) entryId_old(N, m, s, L, d, basis_types, return_type, funcEval, tau, gauss);
elseif L == 0
  entryid_fct = @(N, s, d, basis_types, return_type, funcEval, weights) entryId_quad_new(N, m, s, d, basis_types, return_type, funcEval, tau, weights);
else
  error('m > 0 and L < 0 currently not supported');
end

%nodes for entry id
[nodes_entryid, entryid_fct_params] = entryid_fct(N, 2*s, d, basis_types, 'nodes', [], []);

ffi = cell(d,1);
n_f = cell(d,1);
%samples f for entry identification: d by m by N 
for jj=1:1:d
  basisEval_jj = basis_eval(basis_types, n, nodes_entryid{jj});
  ffi{jj} = basisEval_jj * c;
  n_f{jj} = randn(size(nodes_entryid{jj},1),1);
  if ~isreal(ffi{jj})
    n_f{jj} = n_f{jj} + 1i * randn(size(nodes_entryid{jj},1),1);
  end
end
ffi_a = cell2mat(ffi);
ffi_a = reshape(ffi_a,numel(ffi_a),1);
n_f_a = cell2mat(n_f);
n_f_a = reshape(n_f_a,numel(n_f_a),1);
snorm_f=norm(ffi_a)^2;
snorm_n_f=norm(n_f_a)^2;
res_const=snorm_f/snorm_n_f*10^(-sr/10);
for jj=1:1:d
  samples_entryid{jj} = ffi{jj} + n_f{jj} * sqrt(res_const);
end

ff = cell(d-1,1);
for jj=2:1:d
    %W_j's : cell array of size d-1 by m_1 containing 1 by jj vectors
    w{jj-1} = basis_gen_rnd_nodes(basis_types(1:jj), m_1);
    %Z_j's : cell array of size d-2 by m_2 containing 1 by d-jj vectors 
    z{jj-1} = basis_gen_rnd_nodes(basis_types(jj+1:d), m_2);
    %W_j x Z_j 's: cell array of size m_1m_2 by d-1 containing 1 by d vectors
    wz{jj-1} = zeros(m_1*m_2,d);
    for k=1:m_2
      wz{jj-1}(m_1*(k-1)+1 : m_1*k,:) = [w{jj-1} repmat(z{jj-1}(k,:),m_1,1)];
    end      
end

n_ff = cell(d-1,1);
for jj=2:1:d
  basisEval_jj = basis_eval(basis_types, n, wz{jj-1});
  ff{jj-1} = reshape(basisEval_jj * c, m_1, m_2);
  n_ff{jj-1} = randn(m_1,m_2);
  if ~isreal(ff{jj-1})
    n_ff{jj-1} = n_ff{jj-1} + 1i * randn(m_1,m_2);
  end
end

ff_a = cell2mat(ff);
ff_a = reshape(ff_a,numel(ff_a),1);
snorm_ff=norm(ff_a)^2;
n_ff_a = cell2mat(n_ff);
n_ff_a = reshape(n_ff_a,numel(n_ff_a),1);
snorm_n_ff=norm(n_ff_a)^2;% squared l2 norm of noise
res_const=snorm_ff/snorm_n_ff*10^(-sr/10);
for jj=2:d
  samples_pairing{jj-1} = ff{jj-1} + n_ff{jj-1} * sqrt(res_const);
end
tElapsed_sam = toc(tStart_sam);%stop counting time for sampling


tStart_tot = tic; % Start counting time for main algorithm
%running sublinearized CoSaMP
[a,supp_a,~,num_iter]=sublin_cosamp_v2(samples_ls, nodes_ls, samples_entryid, nodes_entryid, entryid_fct, entryid_fct_params, samples_pairing, w, wz, pairing_fct, N, s, cosamp_max_iter, basis_types);
tElapsed_tot = toc(tStart_tot);% Stop counting time for main algorithm
%comparison of the original coefficients and support with the approximations
aa=[supp_a a];
aa=sortrows(aa);
cc=c;
cc=[n cc];
cc=sortrows(cc);
if size(aa,1) < size(cc,1)
  aa = [aa; zeros(size(cc,1)-size(aa,1),d+1)];
end
diff=aa-cc;
error_n=norm(diff(:,1:d),'fro');% error from support
error_c=norm(diff(:,d+1));% l2 error of the coefficient vector

%deciding if the support is exactly recovered & error_c is within the threshold eps
if abs(error_n-0)>0.5  %|| (error_c>=cosamp_epsilon)
    SF=0;
else
    SF=1;
end

%total sample number
sam_num = numel(samples_ls) + numel(samples_entryid{d});
for jj=1:d-1
  sam_num = sam_num + numel(samples_entryid{jj}) + numel(samples_pairing{jj});
end

rng(cur_rng_state);
